import torch
import numpy as np
import matplotlib
import torchvision
import math
import os
import zipfile
from torch.hub import download_url_to_file
from tqdm import tqdm  # pip3 install tqdm
import matplotlib.pyplot as plt
plt.rcParams["figure.figsize"] = (15, 15)
plt.style.use('dark_background')
import torch.utils.data
import scipy.misc
import scipy.ndimage
import sklearn.decomposition
import argparse # pip3 install argparse
import torch.nn.functional as F

USE_CUDA = torch.cuda.is_available()


class DatasetLungs(torch.utils.data.Dataset):
    def __init__(self, is_train):
        self.class_count = 3

        dataset_name = 'data_lung'
        path_dataset = f'../data/{dataset_name}'
        if not os.path.exists(path_dataset):
            os.makedirs('../data', exist_ok=True)
            download_url_to_file(
                f'http://share.yellowrobot.xyz/1630528570-intro-course-2021-q4/{dataset_name}.zip',
                f'{path_dataset}.zip',
                progress=True
            )
            zipfile.ZipFile(f'{path_dataset}.zip').extractall('../data')

        x = np.load(f'{path_dataset}/x.npy')
        y = np.load(f'{path_dataset}/y.npy')
        self.data = []
        for idx in range(y.shape[0]):
            self.data.append((x[idx], y[idx]))
        if is_train:
            self.data = self.data[:int(len(self)*0.8)]
        else:
            self.data = self.data[int(len(self)*0.8):]

    def normalize(self, data):
        data_max = data.max()
        data_min = data.min()
        if data_min != data_max:
            data = ((data - data_min) / (data_max - data_min))
        return data

    def __len__(self):
        return len(self.data)

    def __getitem__(self, idx):
        x, y = self.data[idx]
        x = self.normalize(x)
        x = torch.FloatTensor(x)
        y = self.normalize(y)
        y = torch.FloatTensor(y)
        return x, y


data_loader_train = torch.utils.data.DataLoader(
    dataset=DatasetLungs(is_train=True),
    batch_size=16,
    shuffle=True,
    drop_last=True
)
data_loader_test = torch.utils.data.DataLoader(
    dataset=DatasetLungs(is_train=False),
    batch_size=16,
    shuffle=False,
    drop_last=True
)


class ResBlock(torch.nn.Module):
    def __init__(self, in_channels, out_channels, stride=1):
        super().__init__()
        self.conv1 = torch.nn.Conv2d(in_channels, out_channels, kernel_size=3, stride=stride, padding=1, bias=False)
        self.gn1 = torch.nn.GroupNorm(num_channels=out_channels, num_groups=math.ceil(out_channels/2))
        self.conv2 = torch.nn.Conv2d(out_channels, out_channels, kernel_size=3, stride=1, padding=1, bias=False)
        self.gn2 = torch.nn.GroupNorm(num_channels=out_channels, num_groups=math.ceil(out_channels/2))
        self.is_bottleneck = False
        if stride != 1 or in_channels != out_channels:
            self.is_bottleneck = True
            self.shortcut = torch.nn.Conv2d(in_channels, out_channels, kernel_size=1, stride=stride, bias=False)

    def forward(self, x):
        residual = x
        out = self.conv1(x)
        out = F.relu(out)
        out = self.gn1(out)
        out = self.conv2(out)
        if self.is_bottleneck:
            residual = self.shortcut(x)
        out += residual
        out = F.relu(out)
        out = self.gn2(out)
        return out

class UNet(torch.nn.Module):
    def __init__(self):
        super().__init__()
        self.maxpool = torch.nn.MaxPool2d(2)
        self.upsample = torch.nn.Upsample(scale_factor=2, mode='bilinear', align_corners=True)

        self.conv_down_1 = ResBlock(1, 8)
        self.conv_down_2 = ResBlock(8, 16)
        self.conv_down_3 = ResBlock(16, 32)
        self.conv_down_4 = ResBlock(32, 64)

        self.conv_middle = ResBlock(64, 64)

        self.conv_up_4 = ResBlock(128, 32)
        self.conv_up_3 = ResBlock(64, 16)
        self.conv_up_2 = ResBlock(32, 8)
        self.conv_up_1 = ResBlock(16, 1)

    def forward(self, x):
        skip_1 = self.conv_down_1.forward(x)
        out = self.maxpool.forward(skip_1)

        skip_2 = self.conv_down_2.forward(out)
        out = self.maxpool.forward(skip_2)

        skip_3 = self.conv_down_3.forward(out)
        out = self.maxpool.forward(skip_3)

        skip_4 = self.conv_down_4.forward(out)

        out = self.conv_middle.forward(skip_4)

        out = torch.cat([out,  skip_4], dim=1)
        out = self.conv_up_4.forward(out)

        out = self.upsample.forward(out)
        out = torch.cat([out,  skip_3], dim=1)
        out = self.conv_up_3.forward(out)

        out = self.upsample.forward(out)
        out = torch.cat([out,  skip_2], dim=1)
        out = self.conv_up_2.forward(out)

        out = self.upsample.forward(out)
        out = torch.cat([out,  skip_1], dim=1)
        out = self.conv_up_1.forward(out)
        return torch.sigmoid(out)


class UNetAdd(torch.nn.Module):
    def __init__(self):
        super().__init__()
        self.maxpool = torch.nn.MaxPool2d(2)
        self.upsample = torch.nn.Upsample(scale_factor=2, mode='bilinear', align_corners=True)

        self.conv_down_1 = ResBlock(1, 8)
        self.conv_down_2 = ResBlock(8, 16)
        self.conv_down_3 = ResBlock(16, 32)
        self.conv_down_4 = ResBlock(32, 64)

        self.conv_middle = ResBlock(64, 64)

        self.conv_up_4 = ResBlock(64, 32)
        self.conv_up_3 = ResBlock(32, 16)
        self.conv_up_2 = ResBlock(16, 8)
        self.conv_up_1 = ResBlock(8, 1)

    def forward(self, x):
        skip_1 = self.conv_down_1.forward(x)
        out = self.maxpool.forward(skip_1)

        skip_2 = self.conv_down_2.forward(out)
        out = self.maxpool.forward(skip_2)

        skip_3 = self.conv_down_3.forward(out)
        out = self.maxpool.forward(skip_3)

        skip_4 = self.conv_down_4.forward(out)

        out = self.conv_middle.forward(skip_4)

        out = out + skip_4
        out = self.conv_up_4.forward(out)

        out = self.upsample.forward(out)
        out = out + skip_3
        out = self.conv_up_3.forward(out)

        out = self.upsample.forward(out)
        out = out + skip_2
        out = self.conv_up_2.forward(out)

        out = self.upsample.forward(out)
        out = out + skip_1
        out = self.conv_up_1.forward(out)
        return torch.sigmoid(out)


def dice_coeficient(predict, target):
    smooth = 1e-5
    intersection = torch.sum(predict * target)
    predict_sum = torch.sum(predict)
    target_sum = torch.sum(target)
    coef = ((2. * intersection + smooth) / (predict_sum + target_sum + smooth))
    return coef


def bce_loss(predict, target, weight = 1):
    predict = predict.view(target.size(0), -1)
    target = target.view(target.size(0), -1)
    loss = -torch.mean(weight*target*torch.log(predict+1e-10) + weight*(1.0 - target)*torch.log(1.0 - predict + 1e-10))
    return loss


def combined_loss(predict, target):
    loss_dice = 1 - dice_coeficient(predict, target)
    loss_bce = bce_loss(predict, target)
    loss = loss_dice + 0.5 * loss_bce
    return loss

model = UNetAdd()
optimizer = torch.optim.Adam(model.parameters(), lr=1e-3)

if USE_CUDA:
    model = model.cuda()

metrics = {}
for stage in ['train', 'test']:
    for metric in [
        'loss',
        'dice_coef',
        'IoU'
    ]:
        metrics[f'{stage}_{metric}'] = []

for epoch in range(1, 100):

    for data_loader in [data_loader_train, data_loader_test]:
        metrics_epoch = {key: [] for key in metrics.keys()}

        stage = 'train'
        if data_loader == data_loader_test:
            stage = 'test'

        for x, y in tqdm(data_loader):

            if USE_CUDA:
                x = x.cuda()
                y = y.cuda()

            if stage == 'test':
                model = model.eval()
                with torch.no_grad():
                    y_prim = model.forward(x)
                    loss_bce = bce_loss(y_prim, y)
                    dice_coef = dice_coeficient(y_prim, y)
                    loss_dice = 1 - dice_coef
                    loss = loss_dice
            else:
                model = model.train()
                y_prim = model.forward(x)
                loss_bce = bce_loss(y_prim, y)
                dice_coef = dice_coeficient(y_prim, y)
                loss_dice = 1 - dice_coef
                loss = loss_dice
            metrics_epoch[f'{stage}_loss'].append(loss.item())
            metrics_epoch[f'{stage}_dice_coef'].append(dice_coef.item())

            if data_loader == data_loader_train:
                loss.backward()
                optimizer.step()
                optimizer.zero_grad()

            loss = loss.cpu()
            y_prim = y_prim.cpu()
            x = x.cpu()
            y = y.cpu()

            np_x = x.data.numpy()
            np_y_prim = y_prim.data.numpy()
            np_y = y.data.numpy()

        metrics_strs = []
        for key in metrics_epoch.keys():
            if stage in key:
                value = np.mean(metrics_epoch[key])
                metrics[key].append(value)
                metrics_strs.append(f'{key}: {round(value, 2)}')
        print(f'epoch: {epoch} {" ".join(metrics_strs)}')

    plt.subplot(121) # row col idx
    plts = []
    c = 0
    for key, value in metrics.items():
        if not np.isnan(value):
            plts += plt.plot(value, f'C{c}', label=key)
            c += 1
    plt.legend(plts, [it.get_label() for it in plts])

    for i, j in enumerate([4, 5, 6, 16, 17, 18]):
        plt.subplot(4, 6, j)
        plt.title('y')
        plt.imshow(np_x[i][0], cmap='Greys', interpolation=None)
        plt.imshow(np_y[i][0], cmap='Reds', alpha=0.5, interpolation=None)
        plt.subplot(4, 6, j+6)
        plt.title('y_prim')
        plt.imshow(np_x[i][0], cmap='Greys', interpolation=None)
        plt.imshow(np.where(np_y_prim[i][0] > 0.8, np_y_prim[i][0], 0), cmap='Reds', alpha=0.5, interpolation=None)

    plt.tight_layout(pad=0.5)
    plt.show()

